
{# documentation.html #}
{% for topic in topics %}
<div class="section">
  <h3 class="section-title">{{ topic.name }}</h3>
  <div class="section-sub-topic-list">
    {% for sub_topic in topic["sub-topics"] %}
    <a class="section-sub-topic" href="{{ sub_topic.link }}">
      <span> {{ sub_topic.name }} </span>
    </a>
    {% endfor %}
  </div>
</div>
{% endfor %}